@@ -136,12 +136,18 @@ typedef struct ambapp_core_s ambapp_core;
 
 struct ambapp_core_s {
 	ambapp_core *next;
-	int index; /* Index of Core */
+	int index; /* Global index of core among all cores */
 #ifdef CONFIG_DEBUG
 	unsigned int vendor;
 	unsigned int device;
 #endif
-	int subindex; /* Index within set of cores with same vendor & device */
+#ifdef HASXML
+	/*
+	 * Index within set of cores with same vendor and device IDs. Only set
+	 * up when using XML.
+	 */
+	int subindex_for_xml;
+#endif
 
 	/* Device Information */
 	ambapp_ahbdev *ahbmst;
