@@ -32,8 +32,8 @@ struct gptimer_timer {
 };
 
 struct gptimer_regs {
-	uint32_t scalar;
-	uint32_t scalar_reload;
+	uint32_t scaler;
+	uint32_t scaler_reload;
 	uint32_t config;
 	uint32_t unused;
 	struct gptimer_timer timer[7];
