@@ -55,6 +55,18 @@ extern char *prop_str_timeout_sec;
 extern char *prop_str_hw_timeout_ms;
 #endif
 
+/* Linux common clock framework properties */
+
+/* Clock producer properties: */
+extern char *prop_str_clock_cells;
+extern char *prop_str_clock_output_names;
+extern char *prop_str_protected_clocks;
+extern char *prop_str_clock_indices;
+
+/* Clock consumer properties */
+extern char *prop_str_clocks;
+extern char *prop_str_clock_names;
+
 extern void opprom_init(void);
 extern void *prom_malloc(int size);
 extern void *prom_calloc(int size);
