@@ -10,6 +10,10 @@
 #include <irqmp.h>
 #include <apbuart.h>
 #include <iommu.h>
+#include <grcg.h>
+#include <gpreg.h>
+
+#define REMOVE_NODE 1
 
 #define GR712RC_SYSTEMID 0x7120e70
 #define GR740_REV0_SYSTEMID 0x07401038
@@ -146,6 +150,15 @@ int gptimer_apb_fixup(ambapp_dev *dev)
 	return 0;
 }
 
+
+static struct gpreg_regs *gpreg = NULL;
+
+void gpreg_init(ambapp_core *core, int level)
+{
+	if (amba_systemid_is_gr740() && !gpreg && core->apbslv)
+		gpreg = (struct gpreg_regs *)core->apbslv->address;
+}
+
 #ifdef CONFIG_WATCHDOG
 
 #define GPTIMER_WATCHDOG_GR740_HW_TIMEOUT 2000
@@ -284,6 +297,396 @@ int canoc_ahb_fixup_virtual(ambapp_dev *dev, void *axb, void *bus, void **pdevp_
 	return 1;
 }
 
+void ambapp_prop_init(struct prop_std *buf, const char *prop_str);
+int ambapp_prop_int(struct prop_std *buf, const char *prop_str, int value);
+int ambapp_prop_str(struct prop_std *buf, const char *prop_str, const char *val_str);
+int ambapp_prop_uintarray(struct prop_std *buf, const char *prop_str,
+			  unsigned int *array, int len);
+int ambapp_prop_str_array(struct prop_std *buf, const char *prop_str,
+			  const char *val_array[]);
+int ambapp_prop_empty(struct prop_std *buf, const char *prop_str);
+int ambapp_prop_intptr(struct prop_std *buf, const char *prop_str, int *value,
+		       int len);
+
+/* =========================================================================*/
+/* Clock gating */
+
+#ifdef CONFIG_CLKGATE
+/**
+ * grcg_add_consumer_props - Common helper function for writing the
+ * clocks and clock-names properties to the provided property buf.
+ *
+ * buf      - The property buffer, where the properties are written.
+ * consumer - The consumer, where the properties will be added to.
+ *
+ * Return values:
+ *  0 = Done. No more properties to be added
+ * >0 = The size of the property struct to be added
+ */
+static int grcg_add_consumer_props(struct prop_std *buf,
+				   struct grcg_consumer *consumer)
+{
+	struct grcg_clkgate *prop_clocks;
+
+	switch (consumer->props_written) {
+	case 0: /* clocks */
+		if (consumer->clk_id >= consumer->producer->num_clks)
+			return 0;
+		prop_clocks = &consumer->producer->clks[consumer->clk_id];
+		return ambapp_prop_intptr(buf, prop_str_clocks,
+					  (int *)prop_clocks,
+					  sizeof(*prop_clocks));
+	case 1: /* clock-names */
+		return ambapp_prop_str(buf, prop_str_clock_names, "gate");
+	default:
+		return 0;
+	}
+}
+
+/**
+ * grcg_add_producer_props - Common function for writing all the supported
+ * clock producer properties that the Linux driver use.
+ *
+ * The following properties are required:
+ * compatible, #clock-cells and clock_output_names
+ *
+ * Optional properties:
+ * protected-clocks and clock-indices
+ *
+ * buf      - The property buffer, where the properties are written.
+ * producer - The producer, where the properties will be added to.
+ *
+ * Return values:
+ *  0 = Done. No more properties to be added
+ * >0 = The size of the property struct to be added
+ */
+static int grcg_add_producer_props(struct prop_std *buf,
+				   struct grcg_producer *producer)
+{
+	switch (producer->props_written) {
+	case 0: /* compatible */
+		return ambapp_prop_str(buf, prop_str_compatible,
+				       "gaisler,grcg");
+	case 1: /* #clock-cells */
+		return ambapp_prop_int(buf, prop_str_clock_cells, 1);
+	case 2: /* clock-output-names */
+		return ambapp_prop_str_array(buf, prop_str_clock_output_names,
+					     producer->clock_output_names);
+	case 3: /* Optional: protected-clocks or clock-indices */
+		if (producer->protected_clks)
+			return ambapp_prop_uintarray(
+				buf, prop_str_protected_clocks,
+				producer->protected_clks,
+				producer->num_protected_clks);
+		else if (producer->clk_indices)
+			return ambapp_prop_uintarray(
+				buf, prop_str_clock_indices,
+				producer->clk_indices,
+				producer->num_clk_indices);
+		break;
+	case 4: /* clock-indices (if protected-clocks was written before) */
+		if (producer->protected_clks && producer->clk_indices)
+			return ambapp_prop_uintarray(
+				buf, prop_str_clock_indices,
+				producer->clk_indices,
+				producer->num_clk_indices);
+		break;
+
+	default:
+		break;
+	}
+
+	return 0;
+}
+
+/* Uses the default property writer for the consumer */
+#define GRCG_CONSUMER_DEFAULT(name, clk_id, device, corenum, producer)         \
+	GRCG_CONSUMER(name, clk_id, device, corenum, producer,                 \
+		      grcg_add_consumer_props)
+
+/* Uses the default property writer for the producer */
+#define GRCG_PRODUCER_DEFAULT(name, corenum)                                   \
+	GRCG_PRODUCER(name, corenum, grcg_add_producer_props)
+
+/* =========================================================================*/
+/* GR740 Clock gating */
+
+/* The prefix used for all GR740 related functions and variables */
+#define GRCG_GR740_PREFIX grcg_gr740
+
+/* GR740 clock output names */
+static const char *grcg_gr740_producer_clk_output_names[] = {
+	"greth0", "greth1",   "spwrtr",	  "pci",     "gr1553b", "grcan",
+	"l4stat", "apbuart0", "apbuart1", "spictrl", "mctrl",	NULL,
+};
+
+/* GR740 - Clock id of the clock gated cores */
+enum grcg_gr740_clkgate_id {
+	GRCG_GR740_GRETH0,
+	GRCG_GR740_GRETH1,
+	GRCG_GR740_SPWRTR,
+	GRCG_GR740_GRPCI2,
+	GRCG_GR740_GR1553B,
+	GRCG_GR740_GRCAN0,
+	GRCG_GR740_GRCAN1,
+	GRCG_GR740_L4STAT,
+	GRCG_GR740_APBUART0,
+	GRCG_GR740_APBUART1,
+	GRCG_GR740_SPICTRL,
+	GRCG_GR740_FTMCTRL,
+	GRCG_GR740_MAX,
+};
+
+/* GR740 - Array of hardware clock gates, one per clock id */
+static struct grcg_clkgate grcg_gr740_producer_clks[GRCG_GR740_MAX] = {
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_GRETH0, 0),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_GRETH1, 1),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_SPWRTR, 2),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_GRPCI2, 3),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_GR1553B, 4),
+	/* Both GRCAN's are controlled by the same gate (5) */
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_GRCAN0, 5),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_GRCAN1, 5),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_L4STAT, 6),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_APBUART0, 7),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_APBUART1, 8),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_SPICTRL, 9),
+	GRCG_CLKGATE_ENTRY(GRCG_GR740_FTMCTRL, 10),
+};
+
+/* No protected clocks for GR740 */
+#define grcg_gr740_producer_protected_clks NULL
+#define grcg_gr740_producer_num_protected_clks 0
+
+/* No clock-indices for GR740 */
+#define grcg_gr740_producer_clk_indices NULL
+#define grcg_gr740_producer_num_clk_indices 0
+
+/* Special property writer for greth1:
+ * greth1 will reference both greth0 and greth1 to work around a known issue.
+ * If greth0 is disabled when greth1 is enabled then that
+ * would also disable greth1. Keeping two references by greth1 will ensure that
+ * the clock never gets disabled as long as greth1 is enabled.
+ */
+static const char *grcg_gr740_greth1_clk_names[] = { "gate", "gate1", NULL };
+static int grcg_gr740_greth1_fixup(struct prop_std *buf,
+				   struct grcg_consumer *consumer)
+{
+	struct grcg_clkgate *prop_clocks;
+
+	switch (consumer->props_written) {
+	case 0:
+		/* Write two phandle entries in the same go so both
+		 * GRCG_GR740_GRETH0 and GRCG_GR740_GRETH1 are referenced by
+		 * greth1
+		 */
+		prop_clocks = &consumer->producer->clks[GRCG_GR740_GRETH0];
+		return ambapp_prop_intptr(buf, prop_str_clocks,
+					  (int *)prop_clocks,
+					  2 * sizeof(*prop_clocks));
+	case 1:
+		return ambapp_prop_str_array(buf, prop_str_clock_names,
+					     grcg_gr740_greth1_clk_names);
+	default:
+		break;
+	}
+
+	return 0;
+}
+
+static struct grcg_producer grcg_gr740_producer0 =
+	GRCG_PRODUCER_DEFAULT(GRCG_GR740_PREFIX, 0);
+
+/* This array contains the clock producer for GR740 */
+static struct grcg_producer *grcg_gr740_producers[] = {
+	&grcg_gr740_producer0,
+};
+
+/**
+ * Helper macro that creates a struct grcg_consumer linked to the producer with
+ * index 0 in the grcg_gr740_producers array
+ */
+#define GRCG_GR740_CONSUMER(clk_id, device, corenum)                           \
+	GRCG_CONSUMER_DEFAULT(GRCG_GR740_PREFIX, clk_id, device, corenum,      \
+			      &grcg_gr740_producer0)
+
+/* This array contains all clock consumers of GR740 */
+static struct grcg_consumer grcg_gr740_consumers[] = {
+	GRCG_GR740_CONSUMER(GRCG_GR740_GRETH0, GAISLER_ETHMAC, 0),
+	GRCG_CONSUMER(GRCG_GR740_PREFIX, GRCG_GR740_GRETH1, GAISLER_ETHMAC,    \
+		      1, &grcg_gr740_producer0, grcg_gr740_greth1_fixup),
+	GRCG_GR740_CONSUMER(GRCG_GR740_SPWRTR, GAISLER_SPWROUTER, 0),
+	GRCG_GR740_CONSUMER(GRCG_GR740_GRPCI2, GAISLER_GRPCI2, 0),
+	GRCG_GR740_CONSUMER(GRCG_GR740_GR1553B, GAISLER_GR1553B, 0),
+	GRCG_GR740_CONSUMER(GRCG_GR740_GRCAN0, GAISLER_GRCAN, 0),
+	GRCG_GR740_CONSUMER(GRCG_GR740_GRCAN1, GAISLER_GRCAN, 1),
+	GRCG_GR740_CONSUMER(GRCG_GR740_L4STAT, GAISLER_L4STAT, 0),
+	GRCG_GR740_CONSUMER(GRCG_GR740_APBUART0, GAISLER_APBUART, 0),
+	GRCG_GR740_CONSUMER(GRCG_GR740_APBUART1, GAISLER_APBUART, 1),
+	GRCG_GR740_CONSUMER(GRCG_GR740_SPICTRL, GAISLER_SPICTRL, 0),
+	GRCG_GR740_CONSUMER(GRCG_GR740_FTMCTRL, GAISLER_FTMCTRL, 0),
+};
+
+/* End GR740 Clock gating */
+/* =========================================================================*/
+
+#endif
+
+/*
+ * Contains grcg clkgating entries
+ */
+static struct grcg grcg_list[] = {
+#ifdef CONFIG_CLKGATE
+	GRCG_ENTRY(GRCG_GR740_PREFIX, amba_systemid_is_gr740),
+#else
+	{ NULL, NULL, 0, NULL, 0 },
+#endif
+};
+
+#define GRCG_NUM GRCG_ARRAY_SIZE(grcg_list)
+
+/**
+ * grcg_is_clkgating_enabled
+ *
+ * Returns 1 if found otherwise 0
+ */
+static int grcg_is_clkgating_enabled(void)
+{
+	int i;
+
+	for (i = 0; i < GRCG_NUM; i++)
+		if (grcg_list[i].is_enabled && grcg_list[i].is_enabled())
+			return 1;
+
+	return 0;
+}
+
+/**
+ * grcg_get_enabled_producer
+ *
+ * Returns the first found (enabled) producer of the specified corenum.
+ * If not found then NULL is returned.
+ */
+static struct grcg_producer *grcg_get_enabled_producer(int corenum)
+{
+	int i, j;
+	struct grcg *grcg;
+
+	for (i = 0; i < GRCG_NUM; i++) {
+		grcg = &grcg_list[i];
+		if (!grcg->is_enabled || !grcg->is_enabled())
+			continue;
+
+		for (j = 0; j < grcg->num_producers; j++)
+			if (grcg->producers[j]->corenum == corenum)
+				return grcg->producers[j];
+	}
+
+	return NULL;
+}
+
+/**
+ * grcg_add_consumer_props_for_core - Adds clocks and clock-names properties
+ *
+ * buf     - The property buffer, where the properties are written.
+ * core    - The core that might support clock gating
+ * corenum - The core subindex or equivalent
+ *           Used when identifing if a core is enabled for clock gating.
+ *
+ * This function is expected to be called from a prop_fixup callback as it
+ * follows the same return value semantics and hence should be called until it
+ * return 0.
+ *
+ * Return values:
+ *  0 = Done. No more properties to be added
+ * >0 = The size of the property struct to be added
+ */
+int grcg_add_consumer_props_for_core(struct prop_std *buf, ambapp_core *core,
+				     int corenum)
+{
+	struct grcg *grcg = NULL;
+	struct grcg_consumer *consumer = NULL;
+	ambapp_dev *dev;
+	int i, j, ret;
+
+	for (i = 0; i < GRCG_NUM; i++) {
+		grcg = &grcg_list[i];
+		if (!grcg->is_enabled || !grcg->is_enabled())
+			continue;
+
+		dev = ambapp_get_first_dev(core);
+
+		/* Check if the core is a clk consumer */
+		for (j = 0; j < grcg->num_consumers; j++) {
+			if (grcg->consumers[j].device == dev->device &&
+			    grcg->consumers[j].corenum == corenum) {
+				consumer = &grcg->consumers[j];
+				break;
+			}
+		}
+
+		/* No match? Then continue with the next enabled GRCG */
+		if (!consumer)
+			continue;
+
+		/* Found a clock consumer, now write the consumer properties */
+		ret = consumer->write_prop(buf, consumer);
+		if (ret)
+			consumer->props_written++;
+
+		return ret;
+	}
+
+	return 0;
+}
+
+int grcg_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	/* Use static as core subindex is not generally available. */
+	static int numgrcgs = -1;
+	struct grcg_producer *producer = NULL;
+	int ret = 0;
+
+	switch (index) {
+	case 0:
+		numgrcgs++;
+		break;
+	default:
+		break;
+	}
+
+	/* Check if this clock gating core is a clk producer */
+	producer = grcg_get_enabled_producer(numgrcgs);
+	if (!producer)
+		return 0;
+
+	/* Found a clock producer, now write the producer properties */
+	ret = producer->write_prop(buf, producer);
+	if (ret)
+		producer->props_written++;
+
+	return ret;
+}
+
+int grcg_fixup_node(ambapp_core *core, struct node *node)
+{
+	/* Use static as core subindex is not generally available. */
+	static int numnode = 0;
+	struct grcg_producer *producer = NULL;
+	int i;
+
+	/* Update phandle's of the clock gates now that the node is available */
+	producer = grcg_get_enabled_producer(numnode);
+	if (producer) {
+		for (i = 0; i < producer->num_clks; i++)
+			producer->clks[i].phandle = (int)node;
+	}
+
+	numnode++;
+
+	return 0;
+}
+
 const char *canoc_compatibles[] = {
 	"gaisler,can_oc",
 	"aeroflexgaisler,can_oc",
@@ -347,9 +750,12 @@ const char *spi_compatibles[] = {
 int spi_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
 {
 	int val;
+	/* Use static as core subindex is not generally available. */
+	static int numspi = -1;
 
 	switch (index) {
 	case 0:
+		numspi++;
 		return ambapp_prop_str_array(buf, prop_str_compatible,
 					     spi_compatibles);
 	case 1:
@@ -357,8 +763,13 @@ int spi_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
 		val = ambapp_dev_freq((ambapp_dev *)core->apbslv);
 		return ambapp_prop_int(buf, prop_str_clockfrequency, val);
 	default:
-		return 0;
+		break;
 	}
+
+	if (grcg_is_clkgating_enabled())
+		return grcg_add_consumer_props_for_core(buf, core, numspi);
+
+	return 0;
 }
 
 
@@ -663,13 +1074,13 @@ void gptimer_init(ambapp_core *core, int level)
 #ifndef CONFIG_BOOTLOADER_FREQ
 	{
 		struct ambapp_bus *ainfo = &ambapp_plb;
-		unsigned long timer_freq, scalar;
+		unsigned long timer_freq, scaler;
 		struct apb_bus *apbbus;
 
-		scalar = BYPASS_LOAD_PA(&gptimer->scalar_reload);
-		if (scalar == 0)
+		scaler = BYPASS_LOAD_PA(&gptimer->scaler_reload);
+		if (scaler == 0)
 			while (1) ; /* Frequency never initialized */
-		timer_freq = int_mul(scalar + 1, 1000000);
+		timer_freq = int_mul(scaler + 1, 1000000);
 
 		/* Set the frequency for the AHB bus that GPTIMER is bridged
 		 * from. The APB frequency is always the same as the AHB
@@ -681,6 +1092,103 @@ void gptimer_init(ambapp_core *core, int level)
 #endif
 }
 
+#ifdef CONFIG_PINMUX
+
+/**
+ * pinmux_gr740_allows_dev - Check if a device is allowed due to pin
+ * multiplexing.
+ *
+ * Func    Remove condition
+ * UART0   GPIO_15 == 0
+ * UART1   GPIO_15 == 0
+ * CAN0    GPIO_15 == 0
+ * CAN1    GPIO_15 == 0
+ * SPW_DBG GPIO_15 == 0
+ * GR1553B GPIO_15 == 0
+ * GRETH1  MEM_IFWIDTH == 0 || MEM_IFWIDTH == 1 && PCI_MODE_ENABLE == 1
+ * PCI	   MEM_IFWIDTH == 0 || MEM_IFWIDTH == 1 && PCI_MODE_ENABLE == 0
+ *
+ * Return values:
+ *  1 = Allowed
+ *  0 = Not allowed
+ */
+int pinmux_gr740_allows_dev(ambapp_dev *dev, int num) {
+	uint32_t ret = 1;
+	uint32_t bootstrap;
+
+	if (!amba_systemid_is_gr740() || !gpreg)
+		return ret;
+
+	bootstrap = BYPASS_LOAD_PA(&gpreg->reg);
+	switch (dev->device) {
+	case GAISLER_GRCAN:
+	case GAISLER_APBUART:
+	case GAISLER_GR1553B:
+		/* Bootstrap signal GPIO_15:
+		 * If this signal is LOW then the PROM/IO interface is enabled
+		 * Otherwise the PROM/IO interface pins are routed to their
+		 * alternative functions.
+		 */
+		if ((bootstrap & GPREG_GR740_BOOTSTRAP_GPIO_15) == 0)
+			ret = 0;
+
+		break;
+	case GAISLER_ETHMAC:
+		/* Only applicable to greth1 */
+		if (num != 1)
+			break;
+		/* Remove greth1 if sdram iface is selected
+		 * or if pci mode is enabled
+		 */
+		if ((bootstrap & GPREG_GR740_BOOTSTRAP_MEM_IFWIDTH) == 0)
+			ret = 0;
+		else if (bootstrap & GPREG_GR740_BOOTSTRAP_PCIMODE_EN)
+			ret = 0;
+
+		break;
+	case GAISLER_GRPCI2:
+	case GAISLER_GRPCI2_DMA:
+		/* Remove pci node if sdram iface is selected
+		 * or if pci mode is disabled.
+		 */
+		if ((bootstrap & GPREG_GR740_BOOTSTRAP_MEM_IFWIDTH) == 0)
+			ret = 0;
+		else if ((bootstrap & GPREG_GR740_BOOTSTRAP_PCIMODE_EN) == 0)
+			ret = 0;
+
+		break;
+	default:
+		break;
+	}
+
+	return ret;
+}
+
+/**
+ * pinmux_allows_core - Check if a core is allowed due to pin multiplexing.
+ *
+ * Return values:
+ *  1 = Allowed
+ *  0 = Not allowed
+ */
+static int pinmux_allows_core(ambapp_core *core, int num)
+{
+	int ret = 1;
+	ambapp_dev *dev = ambapp_get_first_dev(core);
+
+	if (amba_systemid_is_gr740())
+		ret = pinmux_gr740_allows_dev(dev, num);
+
+	return ret;
+}
+
+#else
+static int pinmux_allows_core(ambapp_core *core, int num)
+{
+	return 1;
+}
+#endif
+
 static int apbuart_initied = 0;
 
 /* Find first APBUART for console */
@@ -693,6 +1201,37 @@ int apbuart_apb_fixup(ambapp_dev *dev)
 	return 0;
 }
 
+int apbuart_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	/* Use static as core subindex is not generally available. */
+	static int numuarts = -1;
+
+	switch (index) {
+	case 0:
+		numuarts++;
+		break;
+	default:
+		break;
+	}
+
+	if (grcg_is_clkgating_enabled())
+		return grcg_add_consumer_props_for_core(buf, core, numuarts);
+
+	return 0;
+}
+
+int apbuart_fixup_node(ambapp_core *core, struct node *node)
+{
+	static int numnodes = 0;
+	int ret = 0;
+
+	if (!pinmux_allows_core(core, numnodes))
+		ret = REMOVE_NODE;
+
+	numnodes++;
+	return ret;
+}
+
 /* Turn off GRETH RX/TM DMA operations. This should not have to be done here,
  * however during debug (no proper shutdown or reset/boot) the GRETH may still
  * be active. This may lead to DMA overwrites of early Linux initalization
@@ -717,31 +1256,46 @@ void greth_init(ambapp_core *core, int level)
 
 int greth_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
 {
+	/* Use static as core subindex is not generally available. */
+	static int numgreths = -1;
+
 	switch (index) {
 	case 0:
-		/* Must be last, as we fallthrough to signal no more props */
-		if ((amba_systemid == GR740_REV0_SYSTEMID) ||
-		    (amba_systemid == GR740_REV1_SYSTEMID)) {
+		numgreths++;
+		if (amba_systemid_is_gr740()) {
 			/*
 			 * The two greth cores shares MDIO bus but greth0 is
 			 * nevertheless supposed to only use the PHY with addr 1
 			 * and greth1 the PHY with addr 2.
-			 *
-			 * Use static numgreths as core subindex is not
-			 * generally available.
 			 */
 			static char *prop_str_forcephy;
-			static int numgreths = 0;
-			int subindex = numgreths++;
+			int subindex = numgreths;
 			int phyaddr = subindex + 1;
 
 			PROM_STRDUP_ONCE(prop_str_forcephy, "fixed-phy-addr");
 			return ambapp_prop_int(buf, prop_str_forcephy, phyaddr);
 		}
-		/* Fallthrough */
+		break;
 	default:
-		return 0;
+		break;
 	}
+
+	if (grcg_is_clkgating_enabled())
+		return grcg_add_consumer_props_for_core(buf, core, numgreths);
+
+	return 0;
+}
+
+int greth_fixup_node(ambapp_core *core, struct node *node)
+{
+	static int numnodes = 0;
+	int ret = 0;
+
+	if (!pinmux_allows_core(core, numnodes))
+		ret = REMOVE_NODE;
+
+	numnodes++;
+	return ret;
 }
 
 void canmux_init(ambapp_core *core, int level)
@@ -770,6 +1324,166 @@ int grusbdc_fixup_core(ambapp_core *core)
 	return 0;
 }
 
+int grcan_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	/* Use static as core subindex is not generally available. */
+	static int numgrcan = -1;
+
+	switch (index) {
+	case 0:
+		numgrcan++;
+		break;
+	default:
+		break;
+	}
+
+	if (grcg_is_clkgating_enabled())
+		return grcg_add_consumer_props_for_core(buf, core, numgrcan);
+
+	return 0;
+}
+
+int grcan_fixup_node(ambapp_core *core, struct node *node)
+{
+	static int numnodes = 0;
+	int ret = 0;
+
+	if (!pinmux_allows_core(core, numnodes))
+		ret = REMOVE_NODE;
+
+	numnodes++;
+	return ret;
+}
+
+int l4stat_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	/* Use static as core subindex is not generally available. */
+	static int numl4stat = -1;
+
+	switch (index) {
+	case 0:
+		numl4stat++;
+	default:
+		break;
+	}
+
+	if (grcg_is_clkgating_enabled())
+		return grcg_add_consumer_props_for_core(buf, core, numl4stat);
+
+	return 0;
+}
+
+int gr1553b_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	/* Use static as core subindex is not generally available. */
+	static int numgr1553b = -1;
+
+	switch (index) {
+	case 0:
+		numgr1553b++;
+		break;
+	default:
+		break;
+	}
+
+	if (grcg_is_clkgating_enabled())
+		return grcg_add_consumer_props_for_core(buf, core, numgr1553b);
+
+	return 0;
+}
+
+int gr1553b_fixup_node(ambapp_core *core, struct node *node)
+{
+	static int numnodes = 0;
+	int ret = 0;
+
+	if (!pinmux_allows_core(core, numnodes))
+		ret = REMOVE_NODE;
+
+	numnodes++;
+	return ret;
+}
+
+int spwrouter_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	/* Use static as core subindex is not generally available. */
+	static int numspwrtr = -1;
+
+	switch (index) {
+	case 0:
+		numspwrtr++;
+		break;
+	default:
+		break;
+	}
+
+	if (grcg_is_clkgating_enabled())
+		return grcg_add_consumer_props_for_core(buf, core, numspwrtr);
+
+	return 0;
+}
+
+int ftmctrl_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	/* Use static as core subindex is not generally available. */
+	static int numftmctrl = -1;
+
+	switch (index) {
+	case 0:
+		numftmctrl++;
+		break;
+	default:
+		break;
+	}
+
+	if (grcg_is_clkgating_enabled())
+		return grcg_add_consumer_props_for_core(buf, core, numftmctrl);
+
+	return 0;
+}
+
+int grpci2_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
+{
+	static int numgrpci2 = -1;
+
+	switch (index) {
+	case 0:
+		numgrpci2++;
+		break;
+	default:
+		break;
+	}
+
+	if (grcg_is_clkgating_enabled())
+		return grcg_add_consumer_props_for_core(buf, core, numgrpci2);
+
+	return 0;
+}
+
+int grpci2_fixup_node(ambapp_core *core, struct node *node)
+{
+	static int numnodes = 0;
+	int ret = 0;
+
+	if (!pinmux_allows_core(core, numnodes))
+		ret = REMOVE_NODE;
+
+	numnodes++;
+	return ret;
+}
+
+int grpci2_dma_fixup_node(ambapp_core *core, struct node *node)
+{
+	static int numnodes = 0;
+	int ret = 0;
+
+	if (!pinmux_allows_core(core, numnodes))
+		ret = REMOVE_NODE;
+
+	numnodes++;
+	return ret;
+}
+
 struct amba_driver {
 	unsigned short vendor;
 	unsigned short device;
@@ -809,9 +1523,9 @@ struct amba_driver amba_drivers[] = {
  AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3, NULL, NULL, NULL, leon_init, NULL),
  AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3FT, NULL, NULL, NULL, leon_init, NULL),
  AMBAPP_DRV(VENDOR_GAISLER, GAISLER_APBUART, apbuart_apb_fixup, NULL,
-			NULL,NULL, NULL),
+			apbuart_fixup_prop, NULL, apbuart_fixup_node),
  AMBAPP_DRV(VENDOR_GAISLER, GAISLER_ETHMAC, NULL, NULL,     \
-			greth_fixup_prop, greth_init, NULL),
+ 			greth_fixup_prop, greth_init, greth_fixup_node),
  AMBAPP_DRV(VENDOR_GAISLER, GAISLER_I2CMST, NULL, NULL,     \
 			i2c_fixup_prop, NULL, NULL),
  AMBAPP_DRV(VENDOR_GAISLER, GAISLER_SPICTRL, NULL, NULL,     \
@@ -824,7 +1538,25 @@ struct amba_driver amba_drivers[] = {
  AMBAPP_DRV(VENDOR_GAISLER, GAISLER_CANMUX, NULL, NULL,     \
 			NULL, canmux_init, NULL),
  AMBAPP_DRV(VENDOR_GAISLER, GAISLER_USBDC, NULL, grusbdc_fixup_core,     \
-			NULL, NULL, NULL),
+ 			NULL, NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_CLKGATE, NULL, NULL,     \
+			grcg_fixup_prop, NULL, grcg_fixup_node),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GRCAN, NULL, NULL,     \
+			grcan_fixup_prop, NULL, grcan_fixup_node),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_L4STAT, NULL, NULL,     \
+			l4stat_fixup_prop, NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GR1553B, NULL, NULL,     \
+			gr1553b_fixup_prop, NULL, gr1553b_fixup_node),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_SPWROUTER, NULL, NULL,     \
+			spwrouter_fixup_prop, NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_FTMCTRL, NULL, NULL,     \
+			ftmctrl_fixup_prop, NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GRPCI2, NULL, NULL,     \
+ 			grpci2_fixup_prop, NULL, grpci2_fixup_node),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GPREG, NULL, NULL,     \
+			NULL, gpreg_init, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GRPCI2_DMA, NULL, NULL,     \
+			NULL, NULL, grpci2_dma_fixup_node),
 };
 #define AMBAPP_DRV_NUM (sizeof(amba_drivers)/sizeof(struct amba_driver))
 
@@ -2052,6 +2784,18 @@ void ambapp_prop_init(struct prop_std *buf, const char *prop_str)
 	dp->options = PO_DATA | PO_NEXT;
 }
 
+int ambapp_prop_intptr(struct prop_std *buf, const char *prop_str, int *value, int length)
+{
+	struct prop_ptr *dp = (struct prop_ptr *)buf;
+
+	ambapp_prop_init(buf, prop_str);
+	dp->options = PO_PTR | PO_NEXT;
+	dp->length = length;
+	dp->value = value;
+
+	return sizeof(struct prop_ptr) + dp->length;
+}
+
 int ambapp_prop_int(struct prop_std *buf, const char *prop_str, int value)
 {
 	struct prop_data *dp = (struct prop_data *)buf;
